; This definition is overridden in a nal to allow different ports.
; Included here to facilitate testing at top-level.

(define brl-output-port (current-output-port))

; Any object --> string

(define (brl-string obj)
  (cond ((string? obj) obj)
	((number? obj) (number->string obj))
	(else (call-with-output-string
	       (lambda (p) (display obj p))))))

;
; A BRL template may have an expr (inputs a b c ...) to ease compilation
; This is never the 1st exp in the template.
;

(define (brl-template-inputs l)
  (if (null? l)
      '()
      (if (and (list? (car l))
	       (eq? 'inputs (caar l)))
	  (cdar l)
	  (brl-template-inputs (cdr l)))))

(define (brl-template-inputs-delete! l start)
  (if (or (not (pair? l))
	  (not (pair? (cdr l))))
      start
      (if (and (pair? (cadr l))
	       (eq? 'inputs (caadr l)))
	  (begin
	    (set-cdr! l (cddr l))
	    start)
	  (brl-template-inputs-delete! (cdr l) start))))

;
;  Take a list of expressions (including the inputs expr)
;  and create a named-argument lambda from it.
;

(define (brl-make-nal l)
  (let ((ins (brl-template-inputs l)))
    (cons (eval `(lambda ,(cons 'brl-output-port ins)
		   (letrec-syntax
		       ((brl
			 (syntax-rules (define define-syntax
					 set! set-car! set-cdr!
					 sql-statement-close
					 string-fill! string-set!
					 vector-fill! vector-set!)
			   ((brl (define expr ...)) (define expr ...))
			   ((brl (define-syntax expr ...)) (define-syntax expr ...))
			   ((brl (set! expr ...)) (set! expr ...))
			   ((brl (set! expr ...)) (set! expr ...))
			   ((brl (set-car! expr ...)) (set-car! expr ...))
			   ((brl (set-cdr! expr ...)) (set-cdr! expr ...))
			   ((brl (sql-statement-close expr ...)) (sql-statement-close expr ...))
			   ((brl (string-fill! expr ...)) (string-fill! expr ...))
			   ((brl (string-set! expr ...)) (string-set! expr ...))
			   ((brl (vector-fill! expr ...)) (vector-fill! expr ...))
			   ((brl (vector-set! expr ...)) (vector-set! expr ...))
			   ((brl expr) (display expr brl-output-port))
			   ((brl expr1 expr2 ...) (begin
						    (brl expr1)
						    (brl expr2 ...))))))
		     ,(cons 'brl (brl-template-inputs-delete! l l)))))
	  ins)))

;;;;; Part II: functions to read BRL template files

(define (port-readall p accum)
  (let ((nextobj (read p)))
    (if (eof-object? nextobj)
	accum
	(if (null? accum)
	    (port-readall p (list nextobj))
	    (port-readall p (cons nextobj accum))))))

(define (brl-port-readall p)
  (brl-port-init p)
  (reverse (port-readall p '())))

(define (brl-read-nal p)
  (brl-make-nal (brl-readall p)))

(define (brl-load fname)
  (call-with-input-file fname brl-read-nal))

(define (brl-result nal args)
  (call-with-output-string
   (lambda (p)
     (apply (car nal) (cons p args)))))

;;;;; Part III: binding sets

(define (brl-binding-make name val) (cons name val))

(define (brl-binding-add bset b)
  (let ((existing (assq (car b) bset)))
    (if existing
	(begin
	  (set-cdr! existing
		    ((if (list? (cdr existing))
			 cons
			 list) (cdr b) (cdr existing)))
	  bset)
	(cons b bset))))

(define (brl-bindings bset blist)
  (if (null? blist)
      bset
      (brl-bindings (brl-binding-add bset (car blist))
		    (cdr blist))))

(define (brl-apply nal bset)
  (brl-result nal (map (lambda (sym)
			 (let ((binding (assq sym bset)))
			   (if binding
			       (cdr binding)
			       '())))
		       (cdr nal))))

;
; Rudimentary cache of NALs.  It's a list, not a hash table,
; but should be o.k. for about 100 NALs.
;

(define brl-nal-cache '())

(define (brl-nal-cache-find fname cache)
  (assoc fname cache))

(define (brl-handle-request fname blist)
  (let ((cached (assoc fname brl-nal-cache))
	(new-modtime (file-last-modified fname)))
    (if cached
	(begin
	  (if (not (= new-modtime (caddr cached)))
	      (synchronized
	       cached
	       (begin (set-car! (cdr cached)
				(brl-load fname))
		      (set-car! (cddr cached) new-modtime))))
	  (brl-apply (cadr cached) (brl-bindings '() blist)))
	(begin
	  (synchronized
	   brl-nal-cache
	   (set! brl-nal-cache
		 (cons (list fname (brl-load fname) new-modtime)
		       brl-nal-cache)))
	  (brl-apply (cadar brl-nal-cache)
		     (brl-bindings '() blist))))))

(define (file-last-modified fname)
  (kawa-convert
   ((primitive-virtual-method <java.io.File>
			      "lastModified" <long> ())
    ((primitive-constructor <java.io.File> (<String>))
     fname))))

;;;
;;; JDBC procedures
;;;

(define (sql-driver classname)
  ((primitive-static-method
    <java.lang.Class> "forName" <java.lang.Class> (<String>))
   classname))

(define (sql-connection url . properties)
  (case (length properties)
    ((0) ((primitive-static-method <java.sql.DriverManager> "getConnection"
				   <java.sql.Connection> (<String>))
	  url))
    ((1) ((primitive-static-method <java.sql.DriverManager> "getConnection"
	   <java.sql.Connection> (<String> <java.util.Properties>))
	  url (car properties)))
    ((2) ((primitive-static-method <java.sql.DriverManager> "getConnection"
	   <java.sql.Connection> (<String> <String> <String>))
	  url (car properties) (cadr properties)))))

(define (sql-statement conn)
  ((primitive-interface-method
    <java.sql.Connection> "createStatement" <java.sql.Statement> ())
   conn))

(define (sql-statement-close statement)
  ((primitive-interface-method
    <java.sql.Statement> "close" <void> ())
   statement))

(define (sql-execute-query statement query)
  ((primitive-interface-method
    <java.sql.Statement> "executeQuery" <java.sql.ResultSet> (<String>))
   statement query))

(define (sql-resultset-metadata resultset)
  ((primitive-interface-method
    <java.sql.ResultSet> "getMetaData" <java.sql.ResultSetMetaData> ())
   resultset))

(define (sql-resultset-metadata-columncount rsmd)
  ((primitive-interface-method
    <java.sql.ResultSetMetaData> "getColumnCount" <int> ())
   rsmd))

(define (sql-resultset-metadata-columnlabel rsmd i)
  ((primitive-interface-method
    <java.sql.ResultSetMetaData> "getColumnLabel" <String> (<int>))
   rsmd i))

(define (sql-resultset-metadata-columnname rsmd i)
  ((primitive-interface-method
    <java.sql.ResultSetMetaData> "getColumnName" <String> (<int>))
   rsmd i))

(define (sql-resultset-metadata-columntypename rsmd i)
  ((primitive-interface-method
    <java.sql.ResultSetMetaData> "getColumnTypeName" <String> (<int>))
   rsmd i))

(define (sql-resultset-readnext resultset)
  ((primitive-interface-method
    <java.sql.ResultSet> "next" <boolean> ())
   resultset))

(define (sql-resultset-getobject resultset i)
  ((primitive-interface-method
    <java.sql.ResultSet> "getObject" <Object> (<int>))
   resultset i))

;; Convert a Java object, esp. one returned by JDBC, to
;; the appropriate Kawa object.

(define (kawa-convert java-object)
  (cond
   ((or (instance? java-object <java.lang.String>)
	(instance? java-object <java.lang.StringBuffer>))
    ((primitive-constructor <string> (<java.lang.String>))
     (as <java.lang.String> java-object)))
;; java.util.Date methods are mostly deprecated, but that class
;; more closely resembles the date type in 1990s SQL servers.
;; As more SQL servers handle timezones, GregorianCalendar will be better.
;; Code commented out for now so java.util.Date will stay java.util.Date
;   ((instance? java-object <java.util.Date>)
;    (let ((cal ((primitive-constructor <java.util.GregorianCalendar> ()))))
;      ((primitive-virtual-method
;	<java.util.Calendar> "setTime" <void> (<java.util.Date>))
;       cal java-object)
;      cal))
   ((instance? java-object <java.lang.Integer>)
    ((primitive-constructor <integer> (<int>))
     (as <int> java-object)))
   (else java-object)))

; FIXME: other date-related functions would be useful.

(define (sql-resultset-consrow resultset colnum accum)
  (if (< colnum 1)
      accum
      (sql-resultset-consrow resultset (- colnum 1)
			     (cons (kawa-convert
				    (sql-resultset-getobject
				     resultset colnum))
				   accum))))

(define (sql-resultset-nextrow resultset colcount)
  (if (sql-resultset-readnext resultset)
      (sql-resultset-consrow resultset colcount '())
      #f))

(define (sql-resultset-consall resultset colcount accum)
  (if (sql-resultset-readnext resultset)
      (sql-resultset-consall resultset colcount
			     (cons (sql-resultset-consrow
				    resultset colcount '())
				   accum))
      accum))

(define (sql-statement-results stmt query)
  (let* ((rs (sql-execute-query stmt query))
	 (rsmd (sql-resultset-metadata rs))
	 (colcount (sql-resultset-metadata-columncount rsmd)))
    (reverse (sql-resultset-consall rs colcount '()))))

(define-syntax sql-repeat
  (syntax-rules ()
    ((sql-repeat statement args (query ...) expr ...)
     (let* ((sql-results
	     (sql-execute-query statement
			  (apply string-append (map brl-string
						    (list query ...)))))
	    (sql-metadata (sql-resultset-metadata sql-results))
	    (sql-columncount (sql-resultset-metadata-columncount sql-metadata)))
       (do ((sql-row (sql-resultset-nextrow sql-results sql-columncount)
		     (sql-resultset-nextrow sql-results sql-columncount))
	    (sql-rowcount 1 (+ sql-rowcount 1)))
	   ((not sql-row) (- sql-rowcount 1))
	 (apply (lambda args expr ...) sql-row))))))

;;;
;;; Site-specific definitions
;;;

(if (file-exists? "/home/blewis/sitedef.scm")
    (load "/home/blewis/sitedef.scm"))
